﻿using System.Collections.Generic;
using HeuristicLab.Algorithms.PushGP.Data;
using HeuristicLab.Algorithms.PushGP.Expressions;
using HeuristicLab.Algorithms.PushGP.Stack;

namespace HeuristicLab.Algorithms.PushGP.Interpreter
{
    public class PushGPInterpreter : IInterpreterService
    {
        private Parser parser = new Parser();

        public PushGPInterpreter(Configuration config = null)
        {
            this.Configuration = config ?? new Configuration();

            this.CodeStack = new PushGPStack<Expression>();
            this.ExecStack = new PushGPStack<Expression>();
            this.NameStack = new PushGPStack<string>();
            this.BooleanStack = new PushGPStack<bool>();
            this.IntegerStack = new PushGPStack<long>();
            this.FloatStack = new PushGPStack<double>();
        }

        public Configuration Configuration { get; }

        public IStack<Expression> CodeStack { get; }
        public IStack<Expression> ExecStack { get; }
        public IStack<string> NameStack { get; }
        public IStack<bool> BooleanStack { get; }
        public IStack<long> IntegerStack { get; }
        public IStack<double> FloatStack { get; }

        public IDictionary<string, Expression> CustomExpressions { get; } = new Dictionary<string, Expression>();

        public void Interprete(string code)
        {
            var program = this.Encode(code);

            this.Interprete(program);
        }

        public void Interprete(Expression program)
        {
            this.ExecStack.Push(program);

            if (this.Configuration.TopLevelPushCode)
            {
                this.CodeStack.Push(program);
            }

            // run top level expression
            this.ExecStack.Pop().Eval(this);

            while (this.ExecStack.Count > 0)
            {
                var expression = this.ExecStack.Pop();
                var isNotCodeOp = !expression.OpCode.IsCodeOp();

                if (isNotCodeOp)
                {
                    this.CodeStack.Push(expression);
                }

                expression.Eval(this);

                //System.Console.WriteLine($"Current: {expression}");
                //this.PrintStacks();
                //System.Console.WriteLine("\n\n\n");

                if (isNotCodeOp)
                {
                    CodeStack.Pop();
                }
            }
        }

        public Expression Encode(string code)
        {
            return this.parser.Parse(code);
        }

        public string Decode(Expression expression)
        {
            return string.Concat(expression.ToString());
        }

        public void PrintStacks()
        {
            this.PrintStack("EXEC", this.ExecStack);
            this.PrintStack("CODE", this.CodeStack);
            this.PrintStack("NAME", this.NameStack);
            this.PrintStack("BOOLEAN", this.BooleanStack);
            this.PrintStack("FLOAT", this.FloatStack);
            this.PrintStack("INTEGER", this.IntegerStack);
        }

        private void PrintStack<T>(string name, IStack<T> stack)
        {
            System.Console.WriteLine($"--- {name} ---\n{stack}\n---------------------------------------------------");
        }
    }
}
